"""High-level agent interfaces and implementations for reinforcement learning environments."""

from decuen.agents._agent import ActorAgent, Agent, AgentSettings, CriticAgent
from decuen.agents.q import QAgent, QAgentSettings

__all__ = [
    "Agent", "AgentSettings", "ActorAgent", "CriticAgent",
    "QAgent", "QAgentSettings",
]



"""Workaround for MyPy issue with not being able to store callable functions properly as instance attributes."""

from typing import Any, Generic, TypeVar

CallableType = TypeVar("CallableType")


class _FunctionProperty(Generic[CallableType]):
    """Wrapper class to provide ability to store functions as class properties."""

    def __get__(self, oself: Any, owner: Any) -> CallableType:
        ...

    def __set__(self, oself: Any, value: CallableType) -> None:
        ...



"""Critic interfaces and implementations for generating and learning critical analysis of actions and states."""

from decuen.critics._critic import (ActionCritic, ActionCriticSettings, Critic,
                                    CriticSettings, StateCritic,
                                    StateCriticSettings)
from decuen.critics._q import QCritic, QCriticSettings
from decuen.critics.dqn import DQNCritic, DQNCriticSettings
from decuen.critics.qtable import QTableCritic, QTableCriticSettings

__all__ = [
    "Critic", "CriticSettings",
    "StateCritic", "StateCriticSettings",
    "ActionCritic", "ActionCriticSettings",
    "QCritic", "QCriticSettings",
    "DQNCritic", "DQNCriticSettings",
    "QTableCritic", "QTableCriticSettings",
]



"""Interfaces for arbitrary reinforcement learning agents."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional

import numpy as np  # type: ignore
from gym.spaces import Discrete, Space  # type: ignore

from decuen.actors._actor import Actor
from decuen.critics._critic import ActionCritic
from decuen.memories._memory import Memory, Transition
from decuen.strategies._strategy import Strategy
from decuen.strategies.rand import RandomStrategy
from decuen.utils import checks
from decuen.utils.context import get_context


@dataclass
class AgentSettings:
    """Basic common hyperparameter settings for all agents."""


class Agent(ABC):
    """High-level reinforcement learning agent abstraction.

    This abstraction implements essentially no logic instead delegating all processing to the respective subclasses
    and acts only as a wrapper and liason to the internal structures. There is some high-level functionality provided
    by this interface in order to streamline the use of agents:
        1. the ability to step based on an environment and update internal states,
        2. the ability to act based on a state, and
        3. the ability to learn from memory.
    """

    state_space: Space
    action_space: Space
    memory: Memory
    settings: AgentSettings

    # Current state of the agent
    _state: Optional[np.ndarray]
    # Action taken at that state
    _action: Optional[np.ndarray]
    # Current agent trajectory
    _trajectory: List[Transition]

    def __init__(self, memory: Memory, settings: AgentSettings) -> None:
        """Initialize a generic agent."""
        self.state_space = get_context().state_space
        self.action_space = get_context().action_space
        self.memory = memory
        self.settings = settings

        self._state = None
        self._action = None
        self._trajectory = []

    def init(self, state: np.ndarray) -> np.ndarray:
        """Initialize an agent at the start of a new episode."""
        return self._step(state, None, None)

    def step(self, state: np.ndarray, reward: float, terminal: bool) -> np.ndarray:
        """Step based on a new state, a terminal state signal, and a reward signal.

        Moves this agent forward in time and calculates transitions based on the state of the agent the last time step
        was called. The reward signal corresponds to the transition that caused the migration to this state and the
        terminal signal corresponds to the currently inputted state.
        """
        return self._step(state, reward, terminal)

    def _step(self, state: np.ndarray, reward: Optional[float], terminal: Optional[bool]) -> np.ndarray:
        checks.check_state(self.state_space, state)

        action = self.act(state)

        # If we have no history in this episode, we still don't have anything to store
        if self._state is None or self._action is None or reward is None or terminal is None:
            self._state = state
            self._action = action
            return action

        # Generate the transition and append it to the trajectory
        transition = Transition(state=self._state, action=self._action,
                                new_state=state, reward=reward, terminal=terminal)
        self._trajectory.append(transition)

        # Store the transition in memory and either reset the state of this agent while storing the trajectory
        # if we reached the end of the episode or just continue normally otherwise
        self.memory.store_transition(transition)
        if terminal:
            self.memory.store_trajectory(self._trajectory)
            self._state = None
            self._action = None
            self._trajectory = []
        else:
            self._state = state
            self._action = action

        return action

    @abstractmethod
    def act(self, state: np.ndarray) -> np.ndarray:
        """Generate an action to perform based on a state."""
        ...

    @abstractmethod
    def learn(self) -> None:
        """Learn or improve this agent from memory."""
        ...


class ActorAgent(Agent):
    """Generic actor agent interface.

    Provides functionality to use different modular actors in an agent instead of coding basics of the actor framework
    from scratch as would be the case using a regular agent.
    """

    actor: Actor

    def __init__(self, memory: Memory, actor: Actor, settings: AgentSettings) -> None:
        """Initialize a generic actor agent."""
        super().__init__(memory, settings)
        self.actor = actor

    def act(self, state: np.ndarray) -> np.ndarray:
        """Generate an action to perform based on a state.

        Directly produces the action generated by the actor in this state.
        """
        checks.check_state(self.state_space, state)

        action = self.actor.act(state)

        checks.check_action(self.action_space, action)
        return action

    def learn(self) -> None:
        """Learn or improve this agent from memory.

        Delegates learning to the actor which learns to generate better actions.
        """
        trajectories = self.memory.replay_trajectories()

        for trajectory in trajectories:
            for transition in trajectory:
                checks.check_transition(self.state_space, self.action_space, transition)

        if trajectories:
            self.actor.learn(trajectories)


class CriticAgent(Agent):
    """Generic critic agent interface.

    Provides functionality to use different modular critics in an agent instead of coding basics of the critic framework
    from scratch as would be the case using a regular agent.
    """

    critic: ActionCritic
    strategy: Strategy

    # TODO: support state critic and action critic
    def __init__(self, memory: Memory, critic: ActionCritic, strategy: Strategy, settings: AgentSettings) -> None:
        """Initialize a generic critic agent."""
        super().__init__(memory, settings)
        self.critic = critic
        self.strategy = strategy

    def act(self, state: np.ndarray) -> np.ndarray:
        """Generate an action to perform based on a state.

        Queries the critic to figure out which action would be the best to perform given the state. This method should
        usually be overriden because the implementation runs through all possible actions and produces the one with the
        highest critic score; however, with more information about the interal representation of the critic, very large
        optimizations could be made.
        """
        # TODO: support sampling mechanism to get around needing the action space to be discrete
        if not isinstance(self.action_space, Discrete):
            raise TypeError("critic agent acting is only supported for discrete action spaces")
        return self.strategy.choose([self.critic.crit(state, action) for action in self.action_space.n])

    def learn(self) -> None:
        """Learn or improve this agent from memory.

        Delegates learning to the critic which learns more accurate state or action values.
        """
        transitions = self.memory.replay_transitions()

        for transition in transitions:
            self._populate_values(transition)
            checks.check_transition(self.state_space, self.action_space, transition)

        if transitions:
            self.critic.learn(transitions)

    def _populate_values(self, transition: Transition) -> None:
        """Populate the critic fields in a transition based on the current critic."""
        # TODO: support state critic and action critic
        transition.state_value = self.critic.crit(transition.state, transition.action)


class ActorCriticAgent(ActorAgent, CriticAgent):
    """Generic actor-critic agent interface.

    Provides functionality to use different modular actors and critic in an agent instead of coding basics of the
    actor-critic framework from scratch as would be the case using a regular agent.
    """

    def __init__(self, memory: Memory, actor: Actor, critic: ActionCritic, settings: AgentSettings) -> None:
        """Initialize a generic actor-critic agent."""
        ActorAgent.__init__(self, memory, actor, settings)
        # Note that strategy does nothing in this case since we are never calling the `act` of the `CriticAgent`
        CriticAgent.__init__(self, memory, critic, RandomStrategy(), settings)
        self.actor = actor
        self.critic = critic

    def act(self, state: np.ndarray) -> np.ndarray:
        """Generate an action to perform based on a state according to the actor."""
        return ActorAgent.act(self, state)

    def learn(self) -> None:
        """Learn or improve this agent from memory.

        Calls learning procedures on both the actor and the critic.
        """
        ActorAgent.learn(self)
        CriticAgent.learn(self)



"""Deep Q-network action critics.

Implements both deep Q-learning [1, 2] and double deep Q-learning [3] algorithms to train action critics.

[1] Mnih, Volodymyr; et al. (2013). Playing Atari with Deep Reinforcement Learning
    https://arxiv.org/pdf/1312.5602.pdf
[2] Mnih, Volodymyr; et al. (2015). Human-level control through deep reinforcement learning
    https://web.stanford.edu/class/psych209/Readings/MnihEtAlHassibis15NatureControlDeepRL.pdf
[3] van Hasselt, Hado; et al. (2015). Deep Reinforcement Learning with Double Q-learning
    https://arxiv.org/pdf/1509.06461.pdf
"""

from dataclasses import dataclass
from typing import MutableSequence

import numpy as np  # type: ignore
from gym.spaces import Discrete  # type: ignore
from tensorflow.keras import Sequential  # type: ignore
from tensorflow.keras.models import clone_model  # type: ignore

from decuen.critics._q import QCritic, QCriticSettings
from decuen.memories._memory import Transition
from decuen.utils import checks


@dataclass
class DQNCriticSettings(QCriticSettings):
    """Settings for Q-table critics."""


class DQNCritic(QCritic):
    """Deep Q-network critic.

    See [1], [2], [3] for more details.
    """

    action_space: Discrete
    settings: DQNCriticSettings
    network: Sequential
    _target_network: Sequential

    def __init__(self, model: Sequential, settings: DQNCriticSettings) -> None:
        """Initialize this generic actor critic interface."""
        super().__init__(settings)

        # TODO: possibly generalize to multi-discrete spaces
        if not isinstance(self.action_space, Discrete):
            raise TypeError("action space for Q-table critic must be discrete")

        self.network = model
        self._target_network = clone_model(model)
        self._target_network.set_weights(self.network.get_weights())

    def learn(self, transitions: MutableSequence[Transition]) -> None:
        """Update internal critic representation based on past transitions."""
        self._learn_step += 1

        if not transitions:
            return
        for transition in transitions:
            checks.check_transition(self.state_space, self.action_space, transition)

        states = np.array([transition.state for transition in transitions])
        new_states = np.array([transition.new_state for transition in transitions])

        values = self.network.predict_on_batch(states).numpy()
        new_values = self.network.predict_on_batch(new_states).numpy()
        target_new_values = self._target_network.predict_on_batch(new_states)

        for i, transition in enumerate(transitions):
            target = transition.reward
            if not transition.terminal:
                if self.settings.double:
                    target += self.settings.discount_factor * new_values[i][np.argmax(target_new_values[i])]
                else:
                    target += self.settings.discount_factor * np.max(target_new_values[i])
            values[i][transition.action] = target

        self.network.train_on_batch(states, values)

        if self._learn_step % self.settings.target_update == 0:
            self._target_network.set_weights(self.network.get_weights())

    def crit(self, state: np.ndarray, action: np.ndarray) -> float:
        """Return the Q-value of taking a specific action in a specific state."""
        checks.check_state(self.state_space, state)
        checks.check_action(self.action_space, action)

        return self.values(state)[action]

    def values(self, state: np.ndarray) -> np.ndarray:
        """Return an array of Q-values of all actions in a specific state."""
        checks.check_state(self.state_space, state)

        return self.network.predict_on_batch(np.array([state]))[0].numpy()



"""Interfaces for arbitrary critics and respective settings."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import MutableSequence

import numpy as np  # type: ignore
from gym import Space  # type: ignore

from decuen.memories._memory import Transition
from decuen.utils.context import get_context


@dataclass
class CriticSettings:
    """Basic common settings for all critics."""

    discount_factor: float


@dataclass
class StateCriticSettings(CriticSettings):
    """Basic common settings for all state critics."""


@dataclass
class ActionCriticSettings(CriticSettings):
    """Basic common settings for all action critics."""


# pylint: disable=too-few-public-methods
class Critic(ABC):
    """Generic abstract critic interface.

    Note that this abstraction is more general that what might normally be viewed as a critic in modern literature: we
    generalize the critic to be able to either generate a state-value or an action-value; as such the interface for
    criticising is delegated to the more specialized interfaces of `StateCritic` and `ActionCritic` and those interfaces
    should be used for subclassing or most other purposes.

    This abstraction provides an interface for one general critic functionality common to both state- and action-value
    based critics: the ability to learn based on past transitions and trajectories to improve critical accuracy.
    """

    state_space: Space
    action_space: Space
    settings: CriticSettings
    _learn_step: int

    def __init__(self, settings: CriticSettings) -> None:
        """Initialize this generic critic interface."""
        self.state_space = get_context().state_space
        self.action_space = get_context().action_space
        self.settings = settings
        self._learn_step = 0

    # TODO: support learning from trajectories
    # XXX: possibly return loss or some other metric?
    @abstractmethod
    def learn(self, transitions: MutableSequence[Transition]) -> None:
        """Update internal critic representation based on past transitions."""
        ...


class StateCritic(Critic):
    """Generic abstract state-value critic interface.

    This abstraction slightly specializes the very general `_Critic` interface to output a state-value and as such
    provides an interface for crticising purely based on a state saying how good or bad a given state is to be in.
    """

    settings: StateCriticSettings

    def __init__(self, settings: StateCriticSettings) -> None:
        """Initialize this generic state critic interface."""
        super().__init__(settings)

    @abstractmethod
    def crit(self, state: np.ndarray) -> float:
        """Return a metric of 'goodness' of a state."""
        ...


class ActionCritic(Critic):
    """Generic abstract action-value critic interface.

    This abstraction slightly specializes the very general `_Critic` interface to output an action-value and as such
    provides an interface for criticising based on an action commited in a particular state saying how good or bad
    taking that action in that state is.
    """

    settings: ActionCriticSettings

    def __init__(self, settings: ActionCriticSettings) -> None:
        """Initialize this generic actor critic interface."""
        super().__init__(settings)

    @abstractmethod
    def crit(self, state: np.ndarray, action: np.ndarray) -> float:
        """Return a metric of 'goodness' of taking an action in a state."""
        ...



"""Q-value based critics.

Contains abstractions for Q-value based critics.
"""

from abc import abstractmethod
from dataclasses import dataclass
from typing import MutableSequence

import numpy as np  # type: ignore

from decuen.critics._critic import ActionCritic, ActionCriticSettings
from decuen.memories._memory import Transition


@dataclass
class QCriticSettings(ActionCriticSettings):
    """Q-value based critic settings."""

    target_update: int
    double: bool


class QCritic(ActionCritic):
    """Abstract Q-value based critic to be used for building implementations of Q-value based critics."""

    def __init__(self, settings: QCriticSettings) -> None:
        """Initialize an abstract Q-value based critic."""
        super().__init__(settings)

    @abstractmethod
    def learn(self, transitions: MutableSequence[Transition]) -> None:
        """Update internal critic representation based on past transitions."""
        ...

    @abstractmethod
    def crit(self, state: np.ndarray, action: np.ndarray) -> float:
        """Return a metric of 'goodness' of taking an action in a state."""
        ...

    @abstractmethod
    def values(self, state: np.ndarray) -> np.ndarray:
        """Return an array of Q-values of all actions in a specific state."""
        ...



"""Implementation of an epsilon-greedy action selection strategy."""

from abc import ABC, abstractmethod
from typing import Callable, ClassVar, Collection, Optional

import numpy as np  # type: ignore

from decuen.strategies._strategy import Strategy
from decuen.strategies.greedy import GreedyStrategy
from decuen.strategies.rand import RandomStrategy
from decuen.utils.function_property import _FunctionProperty


# pylint: disable=too-few-public-methods
class EpsilonDecay(ABC):
    """Technique used to decay epsilon in `EpsilonGreedyStrategy`."""

    @abstractmethod
    def decay(self, value: float) -> float:
        """Decay an epsilon value and return the decayed value."""
        ...


class NoEpsilonDecay(EpsilonDecay):
    """No-op epsilon "decay"."""

    def decay(self, value: float) -> float:
        """Return the inputted value directly with no decay."""
        return value


class FunctionEpsilonDecay(EpsilonDecay):
    """Functional epsilon decay.

    Decays based on a custom decay function.
    """

    func: _FunctionProperty[Callable[[float], float]]

    def __init__(self, func: Callable[[float], float]) -> None:
        """Initialize a functional epsilon decay technique."""
        self.func = func

    def decay(self, value: float) -> float:
        """Return the decayed value according to the decaying function."""
        return self.func(value)


class LinearEpsilonDecay(EpsilonDecay):
    """Linear epsilon decay.

    Decays linearly based on a linear decay factor.
    """

    factor: float

    def __init__(self, factor: float) -> None:
        """Initialize a linear epsilon decay technique."""
        self.factor = factor

    def decay(self, value: float) -> float:
        """Return the value reduced by the linear decay factor."""
        return value - self.factor


class ExponentialEpsilonDecay(EpsilonDecay):
    """Exponential epsilon decay.

    Decays geometrically based on a exponential decay factor.
    """

    factor: float

    def __init__(self, factor: float) -> None:
        """Initialize an exponential epsilon decay technique."""
        # TODO: warn against weird factors (i.e. <= 0 or >= 1)
        self.factor = factor

    def decay(self, value: float) -> float:
        """Return the value multiplied by the exponential decay factor."""
        return value * self.factor


# pylint: disable=too-few-public-methods
class EpsilonGreedyStrategy(Strategy):
    """Epsilon-greedy action selection strategy."""

    greedy: ClassVar[GreedyStrategy] = GreedyStrategy()
    random: ClassVar[RandomStrategy] = RandomStrategy()
    epsilon: float
    min_epsilon: float
    max_epsilon: float
    _decay: EpsilonDecay

    def __init__(self, epsilon: float = 1, max_epsilon: float = 1, min_epsilon: float = 0,
                 decay: Optional[EpsilonDecay] = None) -> None:
        """Initialize an epsilon greedy strategy."""
        super().__init__()
        self.epsilon = epsilon
        self.max_epsilon = max_epsilon
        self.min_epsilon = min_epsilon
        self._decay = decay if decay else NoEpsilonDecay()

    def choose(self, action_values: Collection[float]) -> np.ndarray:
        """Choose an action to perform greedily with probability epsilon, otherwise randomly.

        Decays epsilon according to the decay mechanism after choosing an action.
        """
        action = (self.greedy.choose(action_values) if np.random.rand() > self.epsilon
                  else self.random.choose(action_values))
        self.decay()
        return action

    def decay(self) -> None:
        """Decay the epsilon according to the decaying technique."""
        self.epsilon = max(self._decay.decay(self.epsilon), self.min_epsilon)

    def reset(self) -> None:
        """Reset the epsilon to be the maximum epsilon."""
        self.epsilon = self.max_epsilon



"""Interfaces for arbitrary action selection strategies."""

from abc import ABC, abstractmethod
from typing import Collection

import numpy as np  # type: ignore


# pylint: disable=too-few-public-methods
class Strategy(ABC):
    """Action selection strategy interface.

    Note that this interface currently only support discrete action selection.
    """

    def __init__(self) -> None:
        """Initialize an action selection strategy."""
        ...

    @abstractmethod
    def choose(self, action_values: Collection[float]) -> np.ndarray:
        """Choose an action to perform given the respective action values.

        Assumes that each value is indexed respectively by its action.
        """
        ...



"""Implementation of a random action selection strategy."""

from typing import Collection

import numpy as np  # type: ignore

from decuen.strategies._strategy import Strategy


# pylint: disable=too-few-public-methods
class RandomStrategy(Strategy):
    """Random action selection strategy."""

    def choose(self, action_values: Collection[float]) -> np.ndarray:
        """Choose an action to perform uniformly at random."""
        return np.array(np.random.randint(0, len(action_values)))



"""Interface for arbitrary actor-learners and respective settings."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import MutableSequence, Optional

import numpy as np  # type: ignore

from decuen.dists._distribution import Distribution
from decuen.memories._memory import Trajectory
from decuen.policies._policy import Policy


@dataclass
class ActorSettings:
    """Basic common settings for all actor-learners."""


class Actor(ABC):
    """Generic abstract actor-learner interface.

    This abstraction provides interfaces for the two main functionalities of an actor-learner:
        1. the ability to choose an action to perform given a state, and
        2. the ability to learn based on past transitions and trajectories.
    """

    policy: Policy

    @abstractmethod
    def __init__(self, distribution: Optional[Distribution] = None,
                 settings: ActorSettings = ActorSettings()) -> None:
        """Initialize a generic actor-learner."""
        self.policy = Policy(self._generate_policy_parameters,
                             distribution if distribution else self._choose_action_distribution())

    def act(self, state: np.ndarray) -> np.ndarray:
        """Choose an action to perform based on an environment state."""
        return self.policy.act(state).rvs()

    # TODO: support learning from transitions
    # XXX: possibly return loss or some other metric?
    @abstractmethod
    def learn(self, trajectories: MutableSequence[Trajectory]) -> None:
        """Update policy based on past trajectories."""
        ...

    @abstractmethod
    def _generate_policy_parameters(self, state: np.ndarray) -> np.ndarray:
        """Generate policy parameters on-the-fly based on an environment state."""
        ...

    def _choose_action_distribution(self) -> np.ndarray:
        """Choose an action distribution for the policy depending on the action space."""
        # TODO: implement



"""Strategy interfaces and implementations for selecting actions based on only critic evaluation."""

from decuen.strategies._strategy import Strategy
from decuen.strategies.epsilon import EpsilonGreedyStrategy
from decuen.strategies.greedy import GreedyStrategy
from decuen.strategies.rand import RandomStrategy

__all__ = ["Strategy", "EpsilonGreedyStrategy", "GreedyStrategy", "RandomStrategy"]



"""Q-table action critics.

Implements both Q-learning [1] and Double Q-learning [2] algorithms to train action critics.

[1] Watkins, Christopher (1989), Learning from Delayed Rewards (Ph.D. thesis)
    http://www.cs.rhul.ac.uk/~chrisw/new_thesis.pdf
[2] van Hasselt, Hado (2011), Double Q-learning
    http://papers.nips.cc/paper/3964-double-q-learning.pdf
[3] Wikipedia, Q-learning
    https://en.wikipedia.org/wiki/Q-learning
"""

from dataclasses import dataclass
from typing import MutableSequence

import numpy as np  # type: ignore
from gym.spaces import Discrete  # type: ignore

from decuen.critics._q import QCritic, QCriticSettings
from decuen.memories._memory import Transition
from decuen.utils import checks


@dataclass
class QTableCriticSettings(QCriticSettings):
    """Settings for Q-table critics."""

    learning_rate: float


class QTableCritic(QCritic):
    """Q-table critic based on [1]."""

    state_space: Discrete
    action_space: Discrete
    settings: QTableCriticSettings
    table: np.ndarray

    def __init__(self, settings: QTableCriticSettings) -> None:
        """Initialize this generic actor critic interface."""
        super().__init__(settings)

        # # TODO: possibly generalize to multi-discrete spaces
        if not isinstance(self.state_space, Discrete):
            raise TypeError("state space for Q-table critic must be discrete")
        if not isinstance(self.action_space, Discrete):
            raise TypeError("action space for Q-table critic must be discrete")

        # XXX: possibly experiment with different initialization techniques
        self.table = np.zeros((self.state_space.n, self.action_space.n))

    # TODO: implement target tables
    # TODO: implement double Q-learning
    def learn(self, transitions: MutableSequence[Transition]) -> None:
        """Update internal critic representation based on past transitions.

        Utilizes a simple value iteration update using a weighted average of the old value and the new information.
        Many of the concepts used here are elaborated upon in [3].
        """
        learn_rate = self.settings.learning_rate
        discount = self.settings.discount_factor

        for transition in transitions:
            checks.check_transition(self.state_space, self.action_space, transition)

            prev_value = self.crit(transition.state, transition.action)
            new_values = self.values(transition.new_state)
            new_value = transition.reward + (0 if transition.terminal else discount * np.max(new_values))

            self.table[transition.state][transition.action] = (1 - learn_rate) * prev_value + learn_rate * new_value

    def crit(self, state: np.ndarray, action: np.ndarray) -> float:
        """Return the Q-value of taking a specific action in a specific state."""
        checks.check_state(self.state_space, state)
        checks.check_action(self.action_space, action)

        return self.table[state][action]

    def values(self, state: np.ndarray) -> np.ndarray:
        """Return an array of Q-values of all actions in a specific state."""
        checks.check_state(self.state_space, state)

        return self.table[state]



"""Utilities for managing environmental context."""

import functools
from typing import Callable, ClassVar, List, Optional, TypeVar, Union, cast

from gym import Env, Space  # type: ignore


class EnvironmentalContext:
    """Context manager for environmental spaces.

    Use this to create custom contexts that can subsequently either be used as a context manager to enclose a scope in
    context, or be passed into the universal context for use as the default context. This context is then used by the
    `with_context` function wrapper to inject state and action space environmental context into function that need it.

    For example, we can create a scope-encapsulated context by using an instance of this class as a context manager:
    >>> with EnvironmentalContext(state_space=..., action_space=...) as ctx:
    >>>     ...
    Notice that if either <state_space> or <action_space> is unspecified in the initialization, this context manager
    is considered partial. Functions requiring context will still require the unspecified spaces or else an error will
    be raised.

    We can also directly set an environmental context to be the universal context:
    >>> CTX.set_context(EnvironmentalContext(state_space=..., action_space=...))
    """

    _state_space: Optional[Space]
    _action_space: Optional[Space]
    _prev_contexts: List['EnvironmentalContext']

    def __init__(self, *, state_space: Optional[Space] = None, action_space: Optional[Space] = None) -> None:
        """Initialize an envionmental context with optional state and action spaces."""
        self._state_space = state_space
        self._action_space = action_space
        self._prev_contexts = []

    @property
    def state_space(self) -> Space:
        """Access the state space associated with this environmental context."""
        if self._state_space is None:
            return ValueError("no state space in context")
        return self._state_space

    @state_space.setter
    def state_space(self, space: Space) -> None:
        self._state_space = space

    @property
    def action_space(self) -> Space:
        """Access the action space associated with this environmental context."""
        if self._action_space is None:
            return ValueError("no action space in context")
        return self._action_space

    @action_space.setter
    def action_space(self, space: Space) -> None:
        self._action_space = space

    def __enter__(self):
        """Use this environmental context as a context manager."""
        self._prev_contexts.append(CTX.get_context())
        CTX.set_context(self)

    def __exit__(self, *args):
        """Exit this environmental context manager and restore the previous context."""
        CTX.set_context(self._prev_contexts.pop())


class UniversalContext:
    """Singleton encapsulating the universal environmental context.

    Simply provides an interface to get and set the current context. Note that this management is not thread-safe.
    """

    _context: ClassVar[EnvironmentalContext] = EnvironmentalContext()

    @classmethod
    def get_context(cls) -> EnvironmentalContext:
        """Get the current environmental context."""
        return cls._context

    @classmethod
    def set_context(cls, context: EnvironmentalContext) -> None:
        """Set the current environmental context."""
        cls._context = context


CTX = UniversalContext


def get_context() -> EnvironmentalContext:
    """Get the current universal context."""
    return CTX.get_context()


def set_context(context: Union[EnvironmentalContext, Env]) -> None:
    """Set the current universal context.

    Can set based on an `EnvironmentalContext` object or an OpenAI Gym `Env` object.
    """
    if isinstance(context, Env):
        context = EnvironmentalContext(state_space=context.observation_space, action_space=context.action_space)
    CTX.set_context(context)


FuncType = TypeVar("FuncType", bound=Callable)


def with_context(func: FuncType) -> FuncType:
    """Inject the current environmental context into this function.

    This function was constructed to be used as a decorator to initialization functions that expect the state and action
    space of the environment they are working in. Using this functions requires the last two arguments of the wrapped
    function to respectively be `state_space` and `action_space` and they must be keyword-only arguments.

    Using this function injects the current envrionmental context into the function by way of those keyword arguments.
    For this function to have any effect, the current context must be populated which can be done by ways of the
    universal singleton context `UniversalContext` aliased by `CTX` or by directly using an `EnvironmentalContext`
    context manager to associate a context with a scope.
    """
    state_space_kwarg_name = "state_space"
    action_space_kwarg_name = "action_space"

    num_kwargs = func.__code__.co_kwonlyargcount
    kwarg_names = func.__code__.co_varnames

    if num_kwargs < 2:
        raise ValueError(
            "wrapper {with_context.__name__} requires function to accept at least two keyword-only arguments")
    if kwarg_names[-2] != state_space_kwarg_name:
        raise ValueError(
            "wrapper {with_context.__name__} requires second-last argument to be named '{state_space_kwarg_name}'")
    if kwarg_names[-1] != action_space_kwarg_name:
        raise ValueError(
            "wrapper {with_context.__name__} requires last argument to be named '{action_space_kwarg_name}'")

    @functools.wraps(func)
    def context_wrapper(*args, **kwargs):
        if state_space_kwarg_name not in kwargs:
            kwargs[state_space_kwarg_name] = CTX.get_context().state_space
        if action_space_kwarg_name not in kwargs:
            kwargs[action_space_kwarg_name] = CTX.get_context().action_space
        func(*args, **kwargs)
    return cast(FuncType, context_wrapper)



"""Collection of simple checks and exceptions for use around the library."""

import numpy as np  # type: ignore
from gym.spaces import Space  # type: ignore

from decuen.memories._memory import Transition


class DecuenError(Exception):
    """General error class used to encapsulate all errors produced by the library."""


class MalformedStateError(DecuenError):
    """Error raised when a state is found to not belong to an appropriate state space."""


class MalformedActionError(DecuenError):
    """Error raised when an action is found to not belong to an appropriate action space."""


def check_state(state_space: Space, state: np.ndarray) -> None:
    """Check that a state is an appropriately part of a state space.

    Raises a `MalformedStateError` if the state is malformed, i.e not part of the state space.
    """
    if state not in state_space:
        raise MalformedStateError(f"state `{state}` is not in the agent state space `{state_space}`")


def check_action(action_space: Space, action: np.ndarray) -> None:
    """Check that an action is an appropriate part of an action space.

    Raises a `MalformedActionError` if the action is malformed, i.e not part of the action space.
    """
    if action not in action_space:
        raise MalformedActionError(f"action `{action}` is not in the agent state space `{action_space}`")


def check_transition(state_space: Space, action_space: Space, transition: Transition) -> None:
    """Check that a transition is appropriately formed according to given state and action spaces.

    Raises a `MalformedStateError` or `MalformedActionError` if any state or action in the transition is malformed.
    """
    check_state(state_space, transition.state)
    check_action(action_space, transition.action)
    check_state(state_space, transition.new_state)



"""Interfaces and implementations of different distributions for use in policies.

Most of the implementations here are thin wrappers around the appropriate `scipy` distributions used just to create
a common and more manageable interface for use within the rest of the library.
"""



"""Implementation of a possibly bounded uniform experience replay manager."""

import random
from typing import List, Optional

from decuen.memories._memory import Memory, Trajectory, Transition


class UniformMemory(Memory):
    """Sized uniform memory mechanism, stores memories up to a maximum amount if specified."""

    _transitions_cap: Optional[int]
    _trajectories_cap: Optional[int]

    def __init__(self, transition_replay_num: int = 1, trajectory_replay_num: int = 1,
                 transitions_cap: Optional[int] = None, trajectories_cap: Optional[int] = None) -> None:
        """Initialize a uniform memory mechanism."""
        super().__init__([], [], transition_replay_num, trajectory_replay_num)
        self._transitions_cap = transitions_cap
        self._trajectories_cap = trajectories_cap

    def store_transition(self, transition: Transition) -> None:
        """Store a transition in this memory mechanism's buffer with any needed associated information."""
        self.transition = transition
        if self._transitions_cap is not None and len(self._transition_buffer) == self._transitions_cap:
            self._transition_buffer.pop(0)
        self._transition_buffer.append(transition)

    def _replay_transitions(self, num: int) -> List[Transition]:
        return random.choices(self._transition_buffer, k=num)

    def store_trajectory(self, trajectory: Trajectory) -> None:
        """Store a trajectory in this memory mechanism's buffer consisting of a sequence of transitions."""
        self.trajectory = trajectory
        if self._trajectories_cap is not None and len(self._trajectory_buffer) == self._trajectories_cap:
            self._trajectory_buffer.pop(0)
        self._trajectory_buffer.append(trajectory)

    def _replay_trajectories(self, num: int) -> List[Trajectory]:
        return random.choices(self._trajectory_buffer, k=num)



"""Interfaces for arbitrary distributions, single- or multi-variate."""

from abc import ABC, abstractmethod

import numpy as np  # type: ignore


class Distribution(ABC):
    """Arbitrary parameterized distribution abstraction belonging to a particular family of distributions.

    This abstraction provides interfaces for high-level functionality common to all distributions and frequently used
    within reinforcement learning literature and consequently the library.
    """

    parameters: np.ndarray

    def __init__(self, parameters: np.ndarray) -> None:
        """Initialize a parameterized distribution."""
        self.parameters = parameters

    @abstractmethod
    def sample(self) -> np.ndarray:
        """Sample at random from the distribution."""
        ...

    @abstractmethod
    def pdf(self, sample: np.ndarray) -> float:
        """Return the probability of observing a sample."""
        ...

    @abstractmethod
    def log_pdf(self, sample: np.ndarray) -> float:
        """Return the log of the probability of observing a sample."""
        ...

    @abstractmethod
    def entropy(self) -> float:
        """Return the differential entropy of this distribution."""
        ...

    @abstractmethod
    def kl_divergence(self, other: 'Distribution') -> float:
        """Return the Kullback-Leibler divergence between this distribution and another one."""
        ...



"""Interface for arbitrary policies for reinforcement learning agents."""

from typing import Callable, Type

import numpy as np  # type: ignore

from decuen.dists._distribution import Distribution
from decuen.utils.function_property import _FunctionProperty


class Policy:
    """Generic policy usually associated with an agent.

    This abstraction provides an interface for the definition of a policy: returning a distribution over actions given
    a state of the environment. Note that for deterministic policies we still output a distribution over actions but
    hat distribution is spiked at the one specific action we would like to output so that only that action is chosen.
    This simulates the deterministic policy in our stochastic policy setting.
    """

    _parameters_factory: _FunctionProperty[Callable[[np.ndarray], np.ndarray]]
    _distribution_factory: Type[Distribution]

    def __init__(self, parameters_factory: Callable[[np.ndarray], np.ndarray],
                 distribution_factory: Type[Distribution]) -> None:
        """Initialize a generic policy."""
        self._parameters_factory = parameters_factory
        self._distribution_factory = distribution_factory

    def __call__(self, state: np.ndarray) -> np.ndarray:
        """Alias calling an instance of this policy as just calling `act`.

        This induces the effect of the policy seeming like a function as it appears in the literature.
        """
        return self.act(state)

    def act(self, state: np.ndarray) -> np.ndarray:
        """Generate a distribution over action to perform given a state."""
        return self._distribution_factory(self._parameters_factory(state))



"""Implementation of a "void" memory manager, a memory manager that does not actually have any memory.

Simulates the capacity for on-policy training.
"""

from typing import List

from decuen.memories._memory import Memory, Trajectory, Transition


class ShortTermMemory(Memory):
    """Short term memory manager, only stores the most recent events and forgets everything else.

    Can be used for agents that do not require the use of any past memories like most on-policy algorithms.
    """

    def __init__(self) -> None:
        """Initialize a short-term memory mechanism."""
        super().__init__([], [])

    # pylint: disable=useless-super-delegation
    def store_transition(self, transition: Transition) -> None:
        """Store nothing in long-term transition memory."""
        super().store_transition(transition)

    def _replay_transitions(self, num: int) -> List[Transition]:
        return [self.transition] if self.transition else []

    # pylint: disable=useless-super-delegation
    def store_trajectory(self, trajectory: Trajectory) -> None:
        """Store nothing in long-term trajectory memory."""
        super().store_trajectory(trajectory)

    def _replay_trajectories(self, num: int = None) -> List[Trajectory]:
        return [self.trajectory] if self.trajectory else []



"""Interface for arbitrary memory mechanisms for reinforcement learning agents."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import MutableSequence, Optional, Sequence

import numpy as np  # type: ignore

from decuen.dists._distribution import Distribution


@dataclass
class Transition:
    """Simple data structure representing a transition from one state to another with associated information."""

    state: np.ndarray
    action: np.ndarray
    new_state: np.ndarray
    reward: float
    terminal: bool

    behavior: Optional[Distribution] = None
    state_value: Optional[float] = None
    action_value: Optional[float] = None


Trajectory = Sequence[Transition]


class Memory(ABC):
    """Generic abstract memory storage and management unit for an agent.

    This abstraction provides interfaces for the main functionalities of a memory mechanism:
        1. the ability to store a state transition with associated information in memory,
        2. the ability to replay experiences based on some internally implemented mechanism,
        3. the ability to store a trajectory of transitions, and
        4. the ability to replay a trajectory of experiences based on some internally implemented mechanism.
    """

    transition: Optional[Transition]
    trajectory: Optional[Trajectory]
    transition_replay_num: int
    trajectory_replay_num: int

    _transition_buffer: MutableSequence[Transition]
    _trajectory_buffer: MutableSequence[Trajectory]

    def __init__(self,
                 transition_buffer: MutableSequence[Transition], trajectory_buffer: MutableSequence[Trajectory],
                 transition_replay_num: int = 1, trajectory_replay_num: int = 1) -> None:
        """Initialize a generic memory mechanism."""
        self.transition = None
        self.trajectory = None
        self.transition_replay_num = transition_replay_num
        self.trajectory_replay_num = trajectory_replay_num
        self._transition_buffer = transition_buffer
        self._trajectory_buffer = trajectory_buffer

    @abstractmethod
    def store_transition(self, transition: Transition) -> None:
        """Store a transition in this memory mechanism's buffer with any needed associated information."""
        self.transition = transition

    def replay_transitions(self, num: Optional[int] = None) -> MutableSequence[Transition]:
        """Replay experiences from our memory buffer based on some mechanism."""
        if num is None:
            return self._replay_transitions(self.transition_replay_num)
        return self._replay_transitions(num)

    @abstractmethod
    def _replay_transitions(self, num: int) -> MutableSequence[Transition]:
        ...

    @abstractmethod
    def store_trajectory(self, trajectory: Trajectory) -> None:
        """Store a trajectory in this memory mechanism's buffer consisting of a sequence of transitions."""
        self.trajectory = trajectory

    def replay_trajectories(self, num: Optional[int] = None) -> MutableSequence[Trajectory]:
        """Replay trajectories from our memory buffer based on some mechanism."""
        if num is None:
            return self._replay_trajectories(self.trajectory_replay_num)
        return self._replay_trajectories(num)

    @abstractmethod
    def _replay_trajectories(self, num: int) -> MutableSequence[Trajectory]:
        ...



"""Implementation of a greedy action selection strategy."""

from typing import Collection

import numpy as np  # type: ignore

from decuen.strategies._strategy import Strategy


# pylint: disable=too-few-public-methods
class GreedyStrategy(Strategy):
    """Greedy action selection strategy."""

    def choose(self, action_values: Collection[float]) -> np.ndarray:
        """Chooses the highest-valued action greedily."""
        return np.array(np.argmax(action_values))



"""Memory manager interfaces and implementations for different techniques of experience replay."""

from decuen.memories._memory import Memory, Trajectory, Transition
from decuen.memories.short_term import ShortTermMemory
from decuen.memories.uniform import UniformMemory

__all__ = [
    "Transition", "Trajectory", "Memory",
    "ShortTermMemory",
    "UniformMemory",
]



"""Q-value based agent."""

from dataclasses import dataclass

import numpy as np  # type: ignore

from decuen.agents._agent import AgentSettings, CriticAgent
from decuen.critics._q import QCritic
from decuen.memories._memory import Memory
from decuen.strategies._strategy import Strategy


@dataclass
class QAgentSettings(AgentSettings):
    """Settings for Q-value critic based agents."""


class QAgent(CriticAgent):
    """Critic agent with only a Q-value based critic guiding action selection."""

    critic: QCritic

    # pylint: disable=too-many-arguments
    def __init__(self, memory: Memory, critic: QCritic, strategy: Strategy, settings: QAgentSettings) -> None:
        """Initialize a Q-value based critic agent."""
        super().__init__(memory, critic, strategy, settings)

    def act(self, state: np.ndarray) -> np.ndarray:
        """Generate an action to perform based on the Q-values of different actions in a state."""
        return self.strategy.choose(self.critic.values(state))
